/*
  Copyright 2014 by Sean Luke
  Licensed under the Apache 2.0 License (found at the end of this file)


  HFA.C and HFA.H
  Sean Luke
  Version 1.0
  May 2014

  See HFA.H for documentation
*/ 

#include "hfa.h"
#include <stdio.h>

/** The start behavior. */
Behavior start = BEHAVIOR("start", NULL, NULL, NULL);

/* The 'start' function for HFAs.  Handled by macros: you should never need to call this. */
void startHFA(Behavior* hfa)
    {
    hfa->done = HFA_FALSE;
    hfa->failed = HFA_FALSE;
    hfa->counter = 0;
	gettimeofday(&(hfa->timer), NULL);
	hfa->goTransitionedTo = NULL;
    hfa->current = &start;
    }

/* The 'stop' function for HFAs.  Handled by macros: you should never need to call this. */
void stopHFA(Behavior* hfa)
    {
    if (hfa->current == NULL)
        {
        printf("WARNING (stopHFA) current if NULL\n");
        }
    else
        {
        if (!(hfa->current->stop == NULL))
            {
            hfa->current->stop(hfa->current);
            }
        hfa->current->parent = NULL;
        }
    }
                        

/* The 'go' function for HFAs.  Handled by macros: you should never need to call this. */
void goHFA(Behavior* hfa)
    {
    // DETERMINE TRANSITION
    Behavior* newBehavior = NULL;
    if (!(hfa->transition == NULL))
        {
        newBehavior = hfa->transition(hfa);
        }
    if (newBehavior == NULL)
        {
        if (hfa->current == &start)
            {
            newBehavior = hfa->initialState;
            }
        else
            {
            newBehavior = hfa->goTransitionedTo;
            }
        }
    hfa->goTransitionedTo = NULL;

    // PERFORM TRANSITION
    if (!(newBehavior == NULL))
        {
        if (!(newBehavior == hfa->current))
            {
            if (hfa->current == NULL)
                {
                printf("WARNING (goHFA) current is NULL\n");
                }
            else
                {
                if (!(hfa->current->stop == NULL))
                    {
                    hfa->current->stop(hfa->current);
                    }
                hfa->current->parent = NULL;
                }
            newBehavior->parent = hfa;
            if (!(newBehavior->start == NULL))
                {
                newBehavior->start(newBehavior);
                }
            hfa->current = newBehavior;
            hfa->done = HFA_FALSE;
            hfa->failed = HFA_FALSE;
            }
        }
    else
        {
        printf("WARNING (goHFA): NULL new behavior\n");
        }
                
    // EXECUTE
    if ((!(hfa->current == NULL)) &&
        (!(hfa->current->go == NULL)))
        {
        hfa->current->go(hfa->current);
        }
                
    // PERFORM GOTRANSITION IF IT EXISTS
    if (!(hfa->goTransition == NULL))
        {
        hfa->goTransition(hfa);
        }
    }

                        
/* Pulses a top-level behavior once. The first time a behavior is pulsed, its start()
    function is called, then go().  Thereafter, each pulse the go() function is called.
    The stop() function is never called.*/
void pulse(Behavior* behavior)
    {
    if (!(behavior->pulsed))
        {
        if (!(behavior->start == NULL))
            {
            behavior->start(behavior);
            }
        }
    if (!(behavior->go == NULL))
        {
        behavior->go(behavior);
        }
    behavior->pulsed = HFA_TRUE;
    }
        
/* Calls stop() on a top-level behavior, and resets it
    so next time it's pulsed, its start() function is called again. */
void reset(Behavior* behavior)
    {
    if (behavior->pulsed)
        {
        if (!(behavior->stop == NULL))
            {
            behavior->stop(behavior);
            }
        behavior->pulsed = HFA_FALSE;
        }
    }
        
/* Called in the go() function of the BEHAVIOR to mark that it wishes the NEWSTATE to be the
	next current state after this BEHAVIOR.  */
void transit(Behavior* behavior, Behavior* newState)
    {
    if (behavior == NULL)
        {
        printf("WARNING (transition): NULL behavior\n");
        }
    else if (newState == NULL)
        {
        printf("WARNING (transition): NULL newState\n");
        }
    else if (behavior->parent != NULL)
        {
        behavior->parent->goTransitionedTo = newState;
        }
    }

// start() function for bumpCounter behavior
void bumpCounterStart(Behavior* behavior)
    {
    if ((!(behavior == NULL)) && (!(behavior->parent == NULL)))
        {
        behavior->parent->counter += 1;
        } 
    }
        
/* Increments the counter in the HFA. */
Behavior bumpCounter = BEHAVIOR("bumpCounter", &bumpCounterStart, NULL, NULL);


// start() function for the resetCounter behavior
void resetCounterStart(Behavior* behavior)
    {
    if ((!(behavior == NULL)) && (!(behavior->parent == NULL)))
        {
        behavior->parent->counter = 0;
        } 
    }
        
        
/* Sets the counter in the HFA to zero. */
Behavior resetCounter = BEHAVIOR("resetCounter", &resetCounterStart, NULL, NULL);

/* Returns the value of the HFA's timer. */
int currentCounter(Behavior* hfa)
    {
    return hfa->counter;
    }

// start() function for the resetTimer behavior
void resetTimerStart(Behavior* behavior)
    {
    if ((!(behavior == NULL)) && (!(behavior->parent == NULL)))
        {
        gettimeofday(&(behavior->parent->timer), NULL);
        } 
    }
        
/* Resets the timer in the HFA to the current time. */
Behavior resetTimer = BEHAVIOR("resetTimer", &resetTimerStart, NULL, NULL);

/* Returns the time (in microseconds) since the last call to resetTimer or the start of the HFA. */
long currentTimer(Behavior* hfa)
    {
    struct timeval current;
    gettimeofday(&current, NULL);
    
    long diff = (long)(current.tv_sec - hfa->timer.tv_sec);
    long udiff = (long)(current.tv_usec - hfa->timer.tv_usec);
    
    return diff * 1000000 + udiff;
    }


#define FLAG_DONE (0)
#define FLAG_FAILED (1)

// internal function, sets flags and possibly propagates them
void setFlag(Behavior* hfa, int flag)
    {
    if (!(hfa->parent == NULL))
        {
        if (flag == FLAG_DONE)
            {
            hfa->parent->done = 1;
            }
        else // flag == FLAG_FAILED
            {
            hfa->parent->failed = 1;
            }
        if (hfa->parent->propagateFlags)
            {
            setFlag(hfa->parent, flag);
            }
        }
    }


// go() function for the done behavior
void doneGo(Behavior* behavior)
    {
    if ((!(behavior==NULL)) && (!(behavior->parent == NULL)))
        {
        behavior->parent->current = &start;
        setFlag(behavior->parent, FLAG_DONE);
        behavior->parent = NULL;
        }
    }

/* Sets the "done" flag in the HFA's parent, then sets the
	current state to "start", so next iteration the HFA will
	transition out of "start". */
Behavior done = BEHAVIOR("done", NULL, NULL, &doneGo);

// go() function for the sayDone behavior
void sayDoneGo(Behavior* behavior)
    {
    if ((!(behavior==NULL)) && (!(behavior->parent == NULL)))
        {
        setFlag(behavior->parent, FLAG_DONE);
        }
    }

/* Sets the "done" flag in the HFA's parent. */
Behavior sayDone = BEHAVIOR("sayDone", NULL, NULL, &sayDoneGo);

/* Returns the value of the "done" flag in the HFA */
HFA_BOOL isDone(Behavior* hfa)
    {
    return hfa->done;
    }
        
        
// go() function for the failed behavior
void failedGo(Behavior* behavior)
    {
    if ((!(behavior==NULL)) && (!(behavior->parent == NULL)))
        {
        behavior->parent->current = &start;
        setFlag(behavior->parent, FLAG_FAILED);
        behavior->parent = NULL;
        }
    }

/* Sets the "failed" flag in the HFA's parent, then sets the
	current state to "start", so next iteration the HFA will
	transition out of "start". */
Behavior failed = BEHAVIOR("failed", NULL, NULL, &failedGo);

// go() function for the sayFailed behavior
void sayfailedGo(Behavior* behavior)
    {
    if ((!(behavior==NULL)) && (!(behavior->parent == NULL)))
        {
        setFlag(behavior->parent, FLAG_FAILED);
        }
    }

/* Sets the "failed" flag in the HFA's parent. */
Behavior sayfailed = BEHAVIOR("sayfailed", NULL, NULL, &sayfailedGo);

/* Returns the value of the "failed" flag in the HFA */
HFA_BOOL isfailed(Behavior* hfa)
    {
    return hfa->failed;
    }


