/*
  Copyright 2014 by Sean Luke
  Licensed under the Apache 2.0 License (found at the end of this file)


  HFA.C and HFA.H
  Sean Luke
  Version 1.0
  May 2014


  A simple package to make it easy to implement hierarchical finite-state
  automata or HFA in the form of Moore machines, for purposes of programming
  agent behaviors.  This package is not as useful for automata which are intended
  for lexing strings.  An HFA is a finite-state automaton whose states correspond
  to basic behaviors (a hard-coded behavior you have constructed) or to other 
  automata.  While HFA contain other HFA, HFA are not permitted to be recursive: 
  that is, you cannot have an HFA which ultimately contains itself.

  Because C does not have dictionaries (hash tables) built-in, this library is
  fairly rudimentary compared to its companion library, HFA.LUA.  Nor does it
  implement all of the features found in GMU Computer Science Technical Report
  GMU-CS-TR-2014-5, "Building a Hierarchical Finite-State Automaton: A Primer".
  But it does enough to enable you to do rudimentary HFA in C without a lot of
  difficulty.  Read that primer before continuing with this documentation.

  What this package does NOT implement (compared to HFA.LUA and the primer):

  1. Targets
  2. The makeTransition function and per-state transition functions



  -- BOOLEAN VALUES --

  The package defines the data type HFA_BOOL, which is just a char.  
  Variables of this type can be set to either HFA_FALSE (0) or to HFA_TRUE (1).  


  -- BEHAVIORS --

  This package defines a single data type called Behavior, which is a struct that
  contains various slots necessary to implement either (1) a basic hard-coded
  behavior or (2) an HFA which holds, as its states, other Behaviors (either
  basic or HFA).  

  The basic stuff found in all behaviors:

  1. A NAME (as a string).  This really only exists for debugging (for now).
  2. A backpointer to a PARENT of the behavior (or to NULL if there is no parent).
  This item may get changed dynamically to various things as the automaton
  is running.  Before any of the functions below (start/stop/go) are called,
  the PARENT will have been set.
  3. A START function.  This function is called when the behavior is started.
  The function takes a single argument: the behavior itself.  The function may
  be NULL. 
  4. A GO function.  This function is called immediately after START, and
  rapidly / repeatedly thereafter.  The function takes a single argument: 
  the behavior itself.  The function may be NULL.
  5. A STOP function.  This function is called when the behavior is stopped.
  The function takes a single argument: the behavior itself.  The function may
  be NULL.
  5. Whether the behavior has yet been PULSED.  This is used by the pulse function.
  This value is an HFA_BOOL.

  Imagine you had a behavior called myBehavior.  Then you could access these as:

  myBehavior.name
  myBehavior.parent
  myBehavior.pulsed
  myBehavior.start
  myBehavior.stop
  myBehavior.go

  A BASIC BEHAVIOR is a hard-coded behavior you have created yourself.  You don't
  have to implement all three functions (start/stop/go).  For example, if you have
  a one-shot behavior (like "kickBall"), you might implement it by simply providing
  the start function (to kick the ball) and setting the other functions to NULL.
  Or if you have a continuous function (like "forward"), you might do this by
  providing a go function (to move the robot forward by an epsilon), and set
  the others to NULL.  Or if you have an asynchronous behavior (like "walk"),
  you might implement the start function to tell the robot to begin walking,
  set the go function to NULL, and set the stop function to tell the robot to stop
  walking.  And so on.

  A utility macro is provided to you to make this easy:

  Behavior BEHAVIOR(char* name, void (*start)(Behavior*), void (*stop)(Behavior*), void (*go)(Behavior*));

  In most cases you'd store the behavior in a global variable with the same name.  So
  if you were creating a behavior called FOO, you'd do so like this:

  Behavior foo = BEHAVIOR("foo", ... )
                
                                      
                                 
  -- Automata --

  An HFA is a behavior associated with a set of SUB-BEHAVIORS in the form of STATES.
  For purposes of this library, a state and a behavior are the same thing.  At any 
  time, an active HFA has a CURRENT STATE (that is, a current sub-behavior).  There 
  is a special initial state called START, which doesn't do anything special at all
  (not to be confused with the start function).  Thus we have (for some HFA called 
  myHFA):

  myHFA.name            (Because an hfa is a behavior)
  myHFA.parent          (Likewise)
  myHFA.pulsed              (Likewise)
  myHFA.start           (Likewise.  Set to startHFA, see below.)
  myHFA.stop            (Likewise.  Set to stopHFA, see below.)
  myHFA.go              (Likewise.  Set to goHFA, see below.)
  myHFA.current         (The current state)

  An HFA is associated with a TRANSITION FUNCTION which takes the HFA as an 
  argument and returns what the current state ought to be at this time.  This is
  typically done by examining the present setting of the current state and some 
  world information to determine the new current state.  More on transition
  functions in a bit, but for the moment, note that the transition functino is
  stored here:

  myHFA.transition

  Because an HFA is a behavior, it has a start, stop and go function stored.  All
  HFA use exactly the same start, stop, and go functions, and so when you make an HFA
  you don't provide them: the HFA set them up themselves.  These special private
  functions are called startHFA(...), stopHFA(...), and goHFA(...).

  The startHFA(...) function initializes the current state of the HFA to the START
  state, among other things.

  The stopHFA(...) function sets the current state of the HFA to NULL, among other
  things.  If the current state wasn't already NULL or the "start" state, then its
  STOP function is also called recursively.

  The goHFA(...) function is where the interesting work is done.  This function first
  calls the transition function to determine the new current state.  If the current
  state is different from the previous one, then the START function is called on the
  new current state, and furthermore if the previous current state wasn't NULL and 
  wasn't "start", then the previous current state's STOP function is also called.
  Regardless, the current state is changed, and GO function is then called on the
  new current state.

  Notice that in goHFA(...) the transition function is called FIRST, followed by 
  the GO function (and possibly START and STOP).  This means that the "start" 
  state typically never has its GO function called, ever.  This is intentional.  The "start"
  state is just meant to be a dummy state which provides us with a way to define
  a transition funtion which, in essence, will specify the initial behavior performed
  by the HFA.  It's defined in the global variable:

  Behavior start = BEHAVIOR("start", NULL, NULL, NULL);

  Also notice that the HFA doesn't actually store any of its sub-behaviors.  They're
  just specified by the transition function based on the current state of the HFA.
 
  HFA are created with a different utility macro than basic hard-coded behaviors.
  
  Behavior HFA(char* name, Behavior* (*transition)(Behavior*));

  Again, when creating an HFA, you'd typically store it in a global variable of the
  same name as the HFA, thus if you were creating an HFA called FOO, you'd say:

  Behavior foo = HFA("foo", ... )



  -- Transition Functions --

  Transition functions for automata can be complex: they subsume all the edges in the
  finite-state automaton.  It's not very convenient for you to have to write a 
  transition function for the entire automaton, but there it is: C isn't Lua. 

  The transition function you pass into your HFA is going to look like an if-else if-...
  statement, for example:

  Behavior foo = ...
  Behavior bar = ...

  Behavior* myTransition(Behavior* hfa) {
    if (hfa->current == &start) return &foo;  // always start with foo
    else if (hfa->current == &foo) {
      if (iHaveReachedTheWall()) return &bar;
      else return &foo;
      }
    else { // if (hfa->current == &bar)
      if (iHaveMadeALeftTurn()) return &foo;
      else return &bar;
      }
    }

  Behavior myHFA = HFA("myHFA", myTransition); 

  You can create behaviors which specify their own transitions.  More specifically,
  if a behavior is the current state, and within that behavior's GO function, the
  function TRANSIT is called (with a state passed in), then this new state
  will be treated as the current state next time.

  For example, if inside bar's go() function we saw the following code:

  if (iHaveMadeALeftTurn()) transit(&bar, &foo)
  else transit(&bar, &bar);

  ... then we could write our transition function a little differently now:

  Behavior* myTransition(Behavior* hfa) {
    if (hfa->current == &start) return &foo;  // always start with foo
    else if (hfa->current == &foo) {
      if (iHaveReachedTheWall()) return &bar;
      else return &foo;
      }
    else  // if (hfa->current == &bar)
      return NULL; // let bar's go() method handle it
    }

  Though you'd never access this, the new behavior passed into TRANSIT will be
  stored here:

  myHFA.goTransitionedTo

  This second feature can be convenient, but it has two drawbacks.  First, it means that
  your behavior is no longer necessarily modular: it can no longer be easily used in 
  different contexts in different HFA.  Second, even if you use this gizmo for all of your
  behaviors, you still have to specify the transition function for the "start" state.
  In general, I do not recommend it.  

  Note that your transition function takes precedence over any call to TRANSIT.   Even if
  your behavior calls TRANSIT in its GO function, if you specify
  a transition for the behavior in your transition function, the transition will be used
  instead.

  If you do not specify a transition function, or your transition function does not 
  have a transition for a specific current state and that behavior returns nothing, then
  no transition is made: the behavior stays the current state.  No warning is issued.



  -- Running the HFA --

  So you've constructed a hierarchy of automata and basic behaviors.  How do you get the
  whole thing going?  With the utility function PULSE(...)

  pulse(behavior)    initially calls START on the behavior, then GO.  Thereafter
  just calls GO on the behavior.
                            
  Typically you'd just set up your HFA and then call pulse(...) on it forever.  If for 
  some reason you need to reset the HFA after calling pulse(...) on it some number of times,
  you can do so with the following function:

  reset(behavior)         calls STOP on the behavior, then resets it so pulse will
  call START on it again in the future.

  Example:

  pulse(&myHFA)
  pulse(&myHFA)
  pulse(&myHFA)
  -- ... and so on, then to reset:
  reset(&myHFA)
  -- we're now ready to begin pulsing again...



  -- Flags --

  HFA come with a variety of gizmos which you will probably find convenient.  Foremost
  are FLAGS, which permit a child HFA to inform its parent HFA that it believes it
  has completed its task ("done") or that it has "failed".  A flag is set by the
  child behavior transitioning to a FLAG BEHAVIOR as its current state.  When this happens,
  the corresponding flag is set in the *parent* (not the child).  The parent's transition
  functions can test for this flag to determine whether they should now transition away from
  the child as the parent's current state, and go somewhere else.  Flags are reset when
  startHFA(...) is called on the parent and also when the parent's goHFA(...) function
  transitions to a new current state.

  There are four flag behaviors available:

  done          raises the "done" flag and immediately transitions to the "start" state
  failed        raises the "failed" flag and immediately transitions to the "start" state
  sayDone       raises the "done" flag
  sayFailed     raises the "failed" flag

  The first two are the most common, since once you're done (or have failed) you probably
  have no further actions you want to do and might as well start over from the top if for
  some reason your parent HFA doesn't act on this flag being raised.  For
  these two, there's no reason to declare a transition since the transition is
  automatic.  For the second two (setDone, setFailed), you need to provide a transition
  of some kind, typically a unilateral one.  These second two are much less common.

  The flags themselves are stored in the parent here (you'll rarely touch this):

  myHFA.done          (HFA_FALSE by default)
  myHFA.failed        (HFA_FALSE by default)

  Flags may be PROPAGATED.  It's a common pattern in building HFA that you see transition
  functions of the form "if DONE has been rased, then transition to DONE".   This is
  essentially saying "If my child thinks he's done, then I'm done too." (likewise "failed").
  Because it's so common, rather than require this transition, an HFA can declare that all
  of its flags should be propagated to its parent.  This means that if its child sets a flag
  in the HFA, the same flag will immediateliy be set in the HFA's parent.  If the parent
  likewise has propagation set, it'll be set in the grandparent too, and so on.  If this
  feels like throwing an exception, you're on the right track.

  You set propagation in your HFA like this:

  myHFA.propagateFlags = HFA_TRUE;



  -- COUNTERS AND TIMERS --

  Each HFA also has one COUNTER and one TIMER.  HFA nested within one another do not share
  these: they have separate counters and timers.

  A COUNTER is just an integer which starts at zero (when the HFA's startHFA(...) function is
  called) and may be increased or reset to zero by certain behaviors which the HFA may
  transition to as its current state.  These behaviors are:

  bumpCounter            increments the counter by 1
  resetCounter           sets the counter to 0

  The current counter value is stored here:

  myHFA.counter

  ... but it's better style to test the current counter value like this:

  currentCounter(&myHFA)

  Testing the value of a counter is useful in a transition function.  Let's say that
  you wanted to go forward for three steps and if you were still not close enough
  to the ball, you transferred to the run state, then go back to forward.  You might say:

  Behavior* fooTransition(Behavior* hfa) {
    if (hfa->current == &start) return &forward;
    else if (hfa->current == &forward) {
      if (closeToBall()) return &kick;
      else return &bumpCounter;
      }
    else if (hfa->current == &kick) return &rotate;
    else if (hfa->current == &rotate) {
      if (ballAhead()) then return &forward;
      else return &rotate;
      }
    else if (hfa->current == &bumpCounter) {
      if (currentCounter(hfa) > 3) return &run;
      else return &forward;
      }
    else if (hfa->current == &run) return &resetCounter;
    else if (hfa->current == &resetCounter) return &forward;
    }
        

  Similarly, a TIMER is an integer which stores a time interval in microseconds.  
  When the HFA's startHFA(...) function is called, the timer is set to the current time.  
  A single behavior can likewise update the timer to the current time:

  resetTimer            updates the timer to the current time

  The current timer value (as a struct timeval) is stored here:

  myHFA.timer

  ... but almost certainly you'd access it using the more useful function:

  currentTimer(&myHFA)

  This gives you the current time, minus the timer value, and thus the number of microseconds which
  have elapsed since the timer was last reset.  Revisiting the example above, imagine you want
  to go forward not for three steps but rather for three seconds, then do a run at the ball.  You
  might say:

  Behavior* fooTransition(Behavior* hfa) {
    if (hfa->current == &start) return &forward;
    else if (hfa->current == &forward) {
      if (closeToBall()) return &kick;
      else if (currentTimer(hfa) > 3 * 1000000) return &run;
      else return &forward;
      }
    else if (hfa->current == &kick) return &rotate;
    else if (hfa->current == &rotate) {
      if (ballAhead()) then return &forward;
      else return &rotate;
      }
    else if (hfa->current == &run) return &resetTimer;
    else if (hfa->current == &resetCounter) return &forward;
    }


*/

#ifndef __HFA_H__
#define __HFA_H__

#include <sys/time.h>


/* Simple boolean type, plus true and false */
typedef char HFA_BOOL;
#define HFA_FALSE (char)0
#define HFA_TRUE (char)1

/// MACROS FOR CREATING BEHAVIORS
/* Make a simple behavior with a name, start, stop, and go functions */
#define BEHAVIOR(name,start,stop,go) { name, start, stop, go, NULL, HFA_FALSE, NULL, 0, {0, 0}, HFA_FALSE, HFA_FALSE, NULL, HFA_FALSE, NULL, NULL, NULL }

/* Make an HFA with a name and a transition function */
#define HFA(name,transition) { name, &startHFA, &stopHFA, &goHFA, NULL, HFA_FALSE, NULL, 0, {0, 0}, HFA_FALSE, HFA_FALSE, NULL, HFA_FALSE, transition, NULL, NULL }

/* Make a simple HFA with a name and an initial state and no transition function
   (we assume your go functions will call transition(...) inside them. */
#define SIMPLE_HFA(name,initialState) { name, &startHFA, &stopHFA, &goHFA, NULL, HFA_FALSE, NULL, 0, {0, 0}, HFA_FALSE, HFA_FALSE, NULL, HFA_FALSE, NULL, initialState, NULL }

/* Behavior struct */
typedef struct _Behavior {

    // Basic Behavior stuff
              
    /* The Behavior's Name (for debugging) */  
    char* name;

    /* start(Behavior* self) -- called to initialize or reset the behavior */
    void (*start)(struct _Behavior*); //  = NULL;

    /* stop(Behavior* self) -- called to shut down the behavior */
    void (*stop)(struct _Behavior*); // = NULL;

    /* go(Behavior* self, Behavior* parent) -- called to pulse the behavior.  parent may be NULL only if self is the root HFA */
    void (*go)(struct _Behavior*); // = NULL;
                
    /* The parent behavior. Set before start() or go() are called */
    struct _Behavior* parent; // = NULL;
                
    /* Has the start() function already been called on this HFA? */
    HFA_BOOL pulsed; // = HFA_FALSE;
                
                
    ///// Extra HFA Stuff
                
    /* The sub-behavior's go() function may have set a transition to go to.  It is set here. */  
    struct _Behavior* goTransitionedTo; // = NULL;
                
    /* A counter */
    int counter; // = 0;
                
    /* A timer */
    struct timeval timer; // = 0;
                
    /* "done" flag */
    HFA_BOOL done; // = HFA_FALSE;
                
    /* "failed" flag */
    HFA_BOOL failed; // = HFA_FALSE;
                
    /* The current subsidiary behavior (state) of an HFA. */
    struct _Behavior* current; // = NULL;

    /* Should we propagate flags? */
    HFA_BOOL propagateFlags; // = HFA_FALSE;
                
    /* transition(Behavior* self) -- called on HFAs to determine the next subsidiary behavior (state). */
    struct _Behavior* (*transition)(struct _Behavior*); // = NULL;

    /* Is an initial state provided? */
    struct _Behavior* initialState; // = NULL;
                
    /* Was our HFA set with a transition function? */
    void (*goTransition)(struct _Behavior*); // = NULL;
                
    } Behavior;

 
/* The start behavior. */
extern Behavior start;

/* Called in the go() function of the BEHAVIOR to mark that it wishes the NEWSTATE to be the
	next current state after this BEHAVIOR.  */
void transit(Behavior* behavior, Behavior* newState);

/* Behavior and HFA creating functions */

/* The 'go' function for HFAs.  Handled by macros: you should never need to call this. */
void goHFA(Behavior* hfa);

/* The 'start' function for HFAs.  Handled by macros: you should never need to call this. */
void startHFA(Behavior* hfa);

/* The 'stop' function for HFAs.  Handled by macros: you should never need to call this. */
void stopHFA(Behavior* hfa);

/* Pulses a top-level behavior once. The first time a behavior is pulsed, its start()
    function is called, then go().  Thereafter, each pulse the go() function is called.
    The stop() function is never called. */
void pulse(Behavior* behavior);

/* Calls stop() on a top-level behavior, and resets it
    so next time it's pulsed, its start() function is called again. */
void reset(Behavior* behavior);

/* Increments the counter in the HFA. */
extern Behavior bumpCounter;

/* Sets the counter in the HFA to zero. */
extern Behavior resetCounter;

/* Returns the value of the HFA's timer. */
int currentCounter(Behavior* hfa);

/* Resets the timer in the HFA to the current time. */
extern Behavior resetTimer;

/* Returns the time (in microseconds) since the last call to resetTimer or the start of the HFA. */
long currentTimer(Behavior* hfa);

/* Sets the "done" flag in the HFA's parent, then sets the
	current state to "start", so next iteration the HFA will
	transition out of "start". */
extern Behavior done;

/* Sets the "done" flag in the HFA's parent. */
extern Behavior sayDone;

/* Returns the value of the "done" flag in the HFA */
HFA_BOOL isDone(Behavior* hfa);

/* Sets the "failed" flag in the HFA's parent, then sets the
	current state to "start", so next iteration the HFA will
	transition out of "start". */
extern Behavior failed;

/* Sets the "failed" flag in the HFA's parent. */
extern Behavior sayFailed;

/* Returns the value of the "failed" flag in the HFA */
HFA_BOOL isFailed(Behavior* hfa);



#endif

